"""
run the test from the src/invesytoolbox directory:
pytest ../tests/test_date_time.py -v
"""
import datetime
from random import randint

import DateTime
import pendulum
from invesytoolbox.itb_date_time import (
    DT_to_date,
    DT_to_dt,
    DT_to_pendulum,
    add_time,
    convert_datetime,
    create_timedelta,
    date_to_DT,
    date_to_dt,
    daterange_from_week,
    dates_from_week,
    day_from_week,
    dt_to_pendulum,
    get_calendar_week,
    get_dateformat,
    get_dow_number,
    get_isocalendar,
    get_monday,
    is_valid_datetime_string,
    last_week_of_year,
    monday_from_week,
    pendulum_to_DT,
    remove_time,
    str_to_date,
    str_to_DT,
    str_to_dt,
    str_to_pendulum,
    unravel_duration,
)
from testdata.date_time import (
    dates_for_checking,
    durations,
    get_dateformat_error_str_part,
    iso8601_tests,
)


def test_get_dateformat():
    for d, f in dates_for_checking.items():
        fmt = f.get('convert_fmt')
        if not f.get('is_valid'):
            print(f'get_dateformat: ignoring {d} (is invalid)')
            continue

        try:
            df = get_dateformat(datestring=d)
        except ValueError as e:
            if get_dateformat_error_str_part in str(e):
                df = str(e)  # for assert check
            else:
                raise ValueError(e)
        try:
            assert df == fmt
            if get_dateformat_error_str_part in df:
                print(f'- {df} (test OK)')
            else:
                pass
        except AssertionError as e:
            print(f'get_dateformat: AssertionError on {d}: {e}')


def check_convert_from_to(func, convert_from, convert_to):
    for k, v in dates_for_checking.items():
        fmt = v.get('convert_fmt')

        if convert_from == 'string':
            value_to_convert = k
            fmts = [fmt]
        else:
            value_to_convert = v.get(convert_from)
            fmts = [None]

        if not value_to_convert:
            continue

        check_dt = v.get(convert_to)
        if convert_from == 'date':
            check_dt = remove_time(date_time=check_dt)

        for f in fmts:
            kwargs = {}
            if f:
                kwargs = {'fmt': f}
            try:
                dt = func(value_to_convert, **kwargs)
            except ValueError as e:
                if not check_dt:
                    continue  # there should be an error, so it's ok
                else:
                    raise ValueError(e)

            # if result is timezone-aware but reference is naive, make comparable
            if (
                isinstance(dt, datetime.datetime)
                and isinstance(check_dt, datetime.datetime)
                and dt.tzinfo is not None
                and check_dt.tzinfo is None
            ):
                check_dt = check_dt.replace(tzinfo=dt.tzinfo)

            # When converting date (no tz) to DateTime, the result timezone
            # depends on the system timezone. Compare only date/time parts.
            if (isinstance(dt, DateTime.DateTime)
                    and isinstance(check_dt, DateTime.DateTime)
                    and convert_from == 'date'):
                assert dt.parts()[:6] == check_dt.parts()[:6]
            else:
                try:
                    assert dt == check_dt
                except AssertionError as e:
                    raise AssertionError(e)


def test_str_to_dt():
    check_convert_from_to(
        func=str_to_dt,
        convert_from='string',
        convert_to='datetime'
    )


def test_str_to_date():
    check_convert_from_to(
        func=str_to_date,
        convert_from='string',
        convert_to='date'
    )


def test_str_to_DT():
    check_convert_from_to(
        func=str_to_DT,
        convert_from='string',
        convert_to='DateTime'
    )


def test_DT_to_dt():
    check_convert_from_to(
        func=DT_to_dt,
        convert_from='DateTime',
        convert_to='datetime'
    )


def test_DT_to_date():
    check_convert_from_to(
        func=DT_to_date,
        convert_from='DateTime',
        convert_to='date'
    )


def test_date_to_dt():
    check_convert_from_to(
        func=date_to_dt,
        convert_from='date',
        convert_to='datetime'
    )


def test_date_to_dt_with_zero_time():
    """Test that H=0 and M=0 are not treated as falsy."""
    d = datetime.date(2023, 5, 15)

    # H=0, M=30 → 00:30 (not midnight)
    r1 = date_to_dt(d, H=0, M=30)
    assert r1 == datetime.datetime(2023, 5, 15, 0, 30)

    # H=10, M=0 → 10:00 (not midnight)
    r2 = date_to_dt(d, H=10, M=0)
    assert r2 == datetime.datetime(2023, 5, 15, 10, 0)

    # H=0, M=0 → explicit midnight
    r3 = date_to_dt(d, H=0, M=0)
    assert r3 == datetime.datetime(2023, 5, 15, 0, 0)


def test_date_to_DT():
    check_convert_from_to(
        func=date_to_DT,
        convert_from='date',
        convert_to='DateTime'
    )


def test_convert_datetime():
    convert_tos = ('string', 'datetime', 'date', 'DateTime')
    converts = []

    for s, v in dates_for_checking.items():
        for k in convert_tos:
            if k == 'string':
                t = s
            else:
                t = v.get(k)

            if not t:
                continue

            converts.append((s, t))

    for date_string, conv_from in converts:
        dates_for_checking_val = dates_for_checking.get(date_string)

        for conv_to in convert_tos:
            check_obj = dates_for_checking_val.get(conv_to)
            convert_fmt = dates_for_checking_val.get('convert_fmt') if isinstance(conv_from, str) else None

            if check_obj is not None:
                for fmt in (convert_fmt, None):

                    try:
                        res_obj = convert_datetime(
                            date=conv_from,
                            convert_to=conv_to,
                            fmt=fmt
                        )
                    except ValueError as e:
                        if 'is not a valid date or date/time string.' in str(e):
                            # Exception in get_dateformat,
                            # validity is tested in the appropriate test function
                            continue
                        elif 'does not match format' in str(e):
                            continue
                        else:
                            raise ValueError(e)

                    if type(conv_from) is datetime.date:  # isinstance does not work
                        check_obj = remove_time(check_obj)

                    # Handle timezone-aware vs naive datetime comparison
                    if (isinstance(res_obj, datetime.datetime)
                            and isinstance(check_obj, datetime.datetime)
                            and res_obj.tzinfo is not None
                            and check_obj.tzinfo is None):
                        check_obj = check_obj.replace(tzinfo=res_obj.tzinfo)

                    # When converting naive datetime/date to DateTime, the
                    # result timezone depends on the system timezone. Compare
                    # only the date/time parts, not the timezone.
                    if (isinstance(res_obj, DateTime.DateTime)
                            and isinstance(check_obj, DateTime.DateTime)
                            and isinstance(conv_from, (datetime.datetime, datetime.date))
                            and not isinstance(conv_from, DateTime.DateTime)
                            and (not isinstance(conv_from, datetime.datetime)
                                 or conv_from.tzinfo is None)):
                        assert res_obj.parts()[:6] == check_obj.parts()[:6], \
                            f'convert from {conv_from} ({type(conv_from)}) to: {conv_to} with format {fmt}'
                    else:
                        assert res_obj == check_obj, f'convert from {conv_from} ({type(conv_from)}) to: {conv_to} with format {fmt}'


def test_get_calendar_week():
    for s, v in dates_for_checking.items():
        correct_week = v.get('week')

        if not correct_week:
            continue

        for some_date in (
            s,
            v.get('date'),
            v.get('datetime'),
            v.get('DateTime')
        ):
            week = get_calendar_week(some_date)

            assert week == correct_week


def test_get_dow_number():
    for s, v in dates_for_checking.items():
        correct_dow_number = v.get('dow_number')

        if not correct_dow_number:
            continue

        for some_date in (
            s,
            v.get('date'),
            v.get('datetime'),
            v.get('DateTime')
        ):
            dow_number = get_dow_number(some_date)

            assert dow_number == correct_dow_number


def test_get_isocalendar():
    for s, v in dates_for_checking.items():
        correct_isocalendar = v.get('isocalendar')

        if not correct_isocalendar:
            continue

        for some_date in (
            s,
            v.get('date'),
            v.get('datetime'),
            v.get('DateTime')
        ):
            iso_calendar = get_isocalendar(some_date)

            assert iso_calendar == correct_isocalendar


def test_is_valid_datetime_string():
    for s, v in dates_for_checking.items():
        is_valid = v.get('is_valid')
        is_not = '' if is_valid else 'not '

        assert is_valid_datetime_string(s) == is_valid, f'{s} is {is_not}a valid date/time!'

    for s, expected in iso8601_tests.items():
        result = is_valid_datetime_string(s)
        assert result == expected, f'ISO8601: {s} should be {"valid" if expected else "invalid"}'


def test_day_from_week():
    for s, v in dates_for_checking.items():
        correct_day = v.get('date')

        if not correct_day:
            continue

        year = correct_day.year
        week = v.get('week')
        dow = v.get('dow_number')

        day = day_from_week(
            year=year,
            week=week,
            weekday=dow
        )

        assert day == correct_day


def test_monday_from_week():
    for s, v in dates_for_checking.items():
        correct_monday = v.get('monday')

        if not correct_monday:
            continue

        year = correct_monday.year
        week = v.get('week')

        monday = monday_from_week(
            year=year,
            week=week
        )

        assert monday == correct_monday


def test_last_week_of_year():
    for s, v in dates_for_checking.items():
        correct_last_week = v.get('last_week')

        if not correct_last_week:
            continue

        last_week = last_week_of_year(v.get('datetime').year)

        assert last_week == correct_last_week


def test_daterange_from_week():
    for s, v in dates_for_checking.items():
        week = v.get('week')

        if not week:
            continue

        correct_monday = v.get('monday')
        correct_sunday = v.get('sunday')
        year = v.get('datetime').year

        daterange = daterange_from_week(
            year=year,
            week=week
        )

        assert daterange == (correct_monday, correct_sunday)


def test_dates_from_week():
    for s, v in dates_for_checking.items():
        week = v.get('week')

        if not week:
            continue

        year = v.get('datetime').year

        monday = v.get('monday')

        correct_dates = [monday]

        for i in range(1, 7):
            correct_dates.append(monday + datetime.timedelta(days=i))

        dates = dates_from_week(
            year=year,
            week=week
        )

        assert dates == correct_dates


def test_remove_time():
    for v in dates_for_checking.values():
        for dt in ('datetime', 'DateTime'):
            assert remove_time(v.get(dt)) == v.get(f'{dt}_notime')


def test_add_time():
    for _ in range(5):
        date_time = [
            randint(1920, 2023),
            randint(1, 12),
            randint(1, 28),
            randint(1, 11),
            randint(1, 29),
            randint(1, 29)
        ]
        add_time_list = [
            randint(1, 11),
            randint(1, 29),
            randint(1, 29)
        ]

        add_time_str = ':'.join([str(n).rjust(2, '0') for n in add_time_list])

        added_time = [
            *date_time[:3],
            date_time[3] + add_time_list[0],
            date_time[4] + add_time_list[1],
            date_time[5] + add_time_list[2]
        ]

        assert datetime.datetime(*added_time) == add_time(
            datetime.datetime(*date_time),
            add_time_str
        )
        assert DateTime.DateTime(*added_time) == add_time(
            DateTime.DateTime(*date_time),
            add_time_str
        )


def test_unravel_duration():
    for s, d in durations.items():
        for returning in ('list', 'dictionary'):
            assert unravel_duration(duration=s, returning=returning) == d.get(returning)


def test_str_to_dt_timezone():
    from zoneinfo import ZoneInfo

    # default timezone is Europe/Vienna
    result = str_to_dt('15.05.2023')
    assert result.tzinfo == ZoneInfo('Europe/Vienna')

    # explicit UTC
    result = str_to_dt('15.05.2023', tz='UTC')
    assert result.tzinfo == ZoneInfo('UTC')
    assert result.year == 2023
    assert result.month == 5
    assert result.day == 15

    # explicit timezone
    result = str_to_dt('15.05.2023 14:30', tz='US/Eastern')
    assert result.tzinfo == ZoneInfo('US/Eastern')
    assert result.hour == 14
    assert result.minute == 30


def test_str_to_DT_timezone():
    # default timezone Europe/Vienna
    result = str_to_DT('15.05.2023')
    assert result.year() == 2023
    assert result.month() == 5
    assert result.day() == 15
    # DateTime stores offset, not timezone name: CET=GMT+1, CEST=GMT+2
    assert result.timezone() in ('GMT+1', 'GMT+2', 'CET', 'CEST')

    # explicit UTC
    result = str_to_DT('15.05.2023', tz='UTC')
    assert result.timezone() in ('UTC', 'GMT+0', 'GMT')


def test_str_to_date_timezone():
    # date is timezone-independent for most cases, but at boundary times
    # the timezone can matter
    result = str_to_date('15.05.2023')
    assert result == datetime.date(2023, 5, 15)

    result = str_to_date('15.05.2023', tz='UTC')
    assert result == datetime.date(2023, 5, 15)


def test_str_to_pendulum_timezone():
    # default timezone Europe/Vienna
    result = str_to_pendulum('15.05.2023', fmt='DD.MM.YYYY')
    assert result.timezone_name == 'Europe/Vienna'

    # explicit UTC
    result = str_to_pendulum('15.05.2023', fmt='DD.MM.YYYY', tz='UTC')
    assert result.timezone_name == 'UTC'


def test_str_to_pendulum():
    # ISO8601 string without explicit format
    result = str_to_pendulum('2023-05-15T14:30:00')
    assert isinstance(result, pendulum.DateTime)
    assert result.year == 2023
    assert result.month == 5
    assert result.day == 15
    assert result.hour == 14
    assert result.minute == 30

    # with explicit format (pendulum uses DD.MM.YYYY, not %d.%m.%Y)
    result = str_to_pendulum('15.05.2023', fmt='DD.MM.YYYY')
    assert isinstance(result, pendulum.DateTime)
    assert result.year == 2023
    assert result.month == 5
    assert result.day == 15


def test_DT_to_pendulum():
    # use UTC timezone to avoid FixedTimezone issues
    DT = DateTime.DateTime('2023/06/15 10:30:00 UTC')
    result = DT_to_pendulum(DT)
    assert isinstance(result, pendulum.DateTime)
    assert result.year == 2023
    assert result.month == 6
    assert result.day == 15
    assert result.hour == 10
    assert result.minute == 30


def test_dt_to_pendulum():
    dt = datetime.datetime(2023, 6, 15, 10, 30)
    result = dt_to_pendulum(dt)
    assert isinstance(result, pendulum.DateTime)
    assert result.year == 2023
    assert result.month == 6
    assert result.day == 15
    assert result.hour == 10
    assert result.minute == 30


def test_pendulum_to_DT():
    pdt = pendulum.datetime(2023, 6, 15, 10, 30, tz='UTC')
    result = pendulum_to_DT(pdt)
    assert isinstance(result, DateTime.DateTime)
    assert result.year() == 2023
    assert result.month() == 6
    assert result.day() == 15
    # pendulum_to_DT converts to the pendulum timezone via .toZone()
    # so UTC input stays UTC
    assert result.minute() == 30


def test_get_monday():
    # week 1 of 2023: Monday is 2023-01-02
    monday = get_monday(2023, 1)
    assert monday == datetime.date(2023, 1, 2)

    # week 52 of 2022: Monday is 2022-12-26
    monday = get_monday(2022, 52)
    assert monday == datetime.date(2022, 12, 26)

    # returning as DateTime
    monday = get_monday(2023, 1, returning='DateTime')
    assert isinstance(monday, DateTime.DateTime)


def test_create_timedelta():
    # datetime timedelta without days
    td = create_timedelta('02:30:00', 'datetime')
    assert isinstance(td, datetime.timedelta)
    assert td == datetime.timedelta(hours=2, minutes=30)

    # pendulum duration
    td = create_timedelta('01:30:00', 'pendulum')
    assert isinstance(td, pendulum.Duration)

    # DateTime returns a number (fraction of day)
    td = create_timedelta('12:00:00', 'DateTime')
    assert td == 0.5  # 12 hours = 0.5 day

    # hours and minutes only
    td = create_timedelta('03:45', 'datetime')
    assert td == datetime.timedelta(hours=3, minutes=45)


if __name__ == '__main__':
    import pytest
    pytest.main([__file__, '-v'])



# coding=utf-8
"""
run the test from the src/invesytoolbox directory:
pytest ../tests/test_security.py -v
"""

import pytest
from invesytoolbox.itb_security import check_password_security, create_key, create_secure_key


def test_create_secure_key():
    key = create_secure_key()
    assert isinstance(key, str)
    assert len(key) > 0

    key = create_secure_key(bits=128)
    assert isinstance(key, str)
    # 128 bits = 16 bytes = 32 hex chars
    assert len(key) == 32

    key = create_secure_key(max_len=10)
    assert len(key) == 10

    key = create_secure_key(max_len=5)
    assert len(key) == 5

    key = create_secure_key(key='my-predefined-key')
    assert key == 'my-predefined-key'

    key = create_secure_key(key='my-predefined-key', max_len=5)
    assert key == 'my-pr'

    # bits not divisible by 8 must raise ValueError
    with pytest.raises(ValueError, match='multiple of 8'):
        create_secure_key(bits=255)
    with pytest.raises(ValueError, match='multiple of 8'):
        create_secure_key(bits=100)


def test_create_key():
    key = create_key(keyLength=15, mode='alphaNum')
    assert len(key) == 15
    assert key.isalnum()

    key = create_key(keyLength=10, mode='numbers')
    assert len(key) == 10
    assert key.isdigit()

    key = create_key(keyLength=10, mode='lowerChars')
    assert len(key) == 10
    assert key.islower()
    assert key.isalpha()

    key = create_key(keyLength=10, mode='upperChars')
    assert len(key) == 10
    assert key.isupper()
    assert key.isalpha()

    key = create_key(keyLength=10, mode='alphaNumLower')
    assert len(key) == 10
    assert key == key.lower()

    key = create_key(keyLength=10, mode='alphaNumUpper')
    assert len(key) == 10
    assert key == key.upper()

    key = create_key(keyLength=20, mode='all')
    assert len(key) == 20


def test_check_password_security():
    errors = check_password_security('SecurePass1!')
    assert errors == []

    errors = check_password_security('short', length=8)
    assert any('minimum password length' in e for e in errors)

    errors = check_password_security('abcdefgh', check_number=True)
    assert any('digit' in e for e in errors)

    errors = check_password_security('abcdefgh', check_capital=True)
    assert any('uppercase' in e for e in errors)

    errors = check_password_security('passwörd', check_ascii=True)
    assert any('ASCII' in e for e in errors)

    errors = check_password_security('abcdefgh')
    assert errors == []

    errors = check_password_security('Abcd1234', check_number=True, check_capital=True)
    assert errors == []

    # check_nonalphanum: password without special character should fail
    errors = check_password_security('Abcd1234', check_nonalphanum=True)
    assert any('non alpha-numeric' in e for e in errors)

    # check_nonalphanum: password with special character should pass
    errors = check_password_security('Abcd123!', check_nonalphanum=True)
    assert not any('non alpha-numeric' in e for e in errors)



# -*- coding: utf-8 -*-
"""
itb_www
=======
"""

__all__ = [
    'ZOPE_DATA_TYPES',
    'change_query_string',
]

import urllib.parse
from typing import Literal, Optional

try:
    from typing import assert_never
except ImportError:
    from typing_extensions import assert_never

ZOPE_DATA_TYPES = {
    'boolean',
    'int',
    'long',
    'float',
    'string',
    'text',
    'list',
    'tuple',
    'tokens',
    'lines',
    ':ignore_empty'
}


def change_query_string(
    *,
    params: dict,
    url: Optional[str] = None,
    query_string: Optional[str] = None,
    delete_remaining: bool = False,
    returning: Literal['url', 'query_string', 'auto'] = 'auto'
) -> str:
    """Change arguments in a query string

    There are no positional arguments, because it must be clear
    if an url or a query string is provided.

    :param params: dictionary of query parameters to set
    :param url: full URL whose query string should be modified
    :param query_string: raw query string to modify (alternative to ``url``)
    :param delete_remaining: only keep arguments specified in params
    :param returning: ``'url'``, ``'query_string'`` or ``'auto'`` (inferred from input)
    :return: modified URL or query string
    """
    if returning == 'auto':
        if url:
            returning = 'url'
        elif query_string:
            returning = 'query_string'

    if url:
        url_parts = urllib.parse.urlparse(url)
        query = dict(urllib.parse.parse_qsl(url_parts.query))
    elif query_string:
        if not url and returning == 'url':
            raise ValueError('an url is needed to return an url (only query string provided)')
        query = dict(urllib.parse.parse_qsl(query_string))
    else:
        raise TypeError('change_query_string needs an "url" or "query_string" argument!')

    if delete_remaining:
        query = params
    else:
        query.update(params)

    new_query_string = urllib.parse.urlencode(query)
    if '%3A' in new_query_string:
        for typ in ZOPE_DATA_TYPES:
            new_query_string = new_query_string.replace(f'%3A{typ}', f':{typ}')

    if returning == 'query_string':
        return new_query_string

    if returning == 'url':
        new_url = url_parts._replace(query=new_query_string).geturl()
        return new_url

    assert_never(returning)  # unreachable



# -*- coding: utf-8 -*-
"""
itb_text_name
=============
"""

__all__ = [
    'NAME_PREFIXES',
    'TRANS_DICT',
    'adjust_spaces_on_punctuation',
    'and_list',
    'capitalize_name',
    'could_be_a_name',
    'get_gender',
    'leet',
    'map_special_chars',
    'normalize_name',
    'normalize_text',
    'sort_names',
    'sort_word_list',
]

import random
import re
import unicodedata
from string import punctuation
from typing import Dict, List, Literal, Optional, Union, cast

import gender_guesser.detector as gender
import unidecode
from nameparser import HumanName

GENDER_DETECTOR = gender.Detector(case_sensitive=False)

TRANS_DICT = {
    'ä': 'ae',
    'Ä': 'Ae',
    'ö': 'oe',
    'Ö': 'Oe',
    'ü': 'ue',
    'Ü': 'Ue',
    'ß': 'ss',
    'æ': 'ae',
    'Æ': 'Ae'
}

CHAR_NB_MAP = {
    'a': ('a', '4', '@'),
    'b': ('b', '6', '8', 'I3', '13', '!3'),
    'c': ('c', '('),
    'd': ('d', ')', '!)', 'cl'),
    'e': ('e', '3', '€'),
    'f': ('f', '7', '/='),
    'g': ('g', '9', '6'),
    'h': ('h', '8', '#', '!-!'),
    'i': ('i', '1', '!'),
    'j': ('j', '1', '_)'),
    'k': ('k', '7'),
    'l': ('1', '!', '(_'),  # intentionally no 'l' — too easily confused with 'I' and '1'
    'm': ('m', 'nn', '(V)', '/VA'),
    'n': ('n', '/V'),
    'o': ('o', '0'),
    'p': ('p', '?'),
    'q': ('q', '9', '2', '()_'),
    'r': ('r', '8', '12'),
    's': ('s', '5', '$', '§'),
    't': ('t', '7', '+'),
    'u': ('u', '(_)'),
    'v': ('v',),
    'w': ('w', 'vv', 'uu'),
    'x': ('x', '><'),
    'y': ('y', '7', '?'),
    'z': ('z', '2', '7_')
}

# regex pattern
LANG_PUNCT_SPACES = {
    'fr': {
        'regex': r'([!?:;%])',
        'narrow': '\u202f\\1',
        'normal': r' \1'
    },
    'es': {
        'regex': r'([¿¡])',
        'narrow': '\\1\u202f',
        'normal': r'\1 '
    },
    'ro': {
        'regex': r'([!?:;])',
        'narrow': '\u202f\\1',
        'normal': r' \1'
    }
}

def adjust_spaces_on_punctuation(
    text: str,
    language: str,
    space: str = 'narrow'
) -> str:
    """Adjust spaces before or after punctuation marks according to the language.
    By default the narrow space (\\u202f) is used for adjustments.

    :param text: input text
    :param language: language code (e.g. ``'fr'``, ``'es'``, ``'ro'``)
    :param space: ``'narrow'`` (default, uses U+202F) or ``'normal'``
    :return: text with adjusted spacing around punctuation
    """
    # Remove existing spaces before punctuation (universal for all languages)
    text = re.sub(r'\s+([!?:;%\),.])', r'\1', text)

    lang_data = LANG_PUNCT_SPACES.get(language)

    if lang_data:
        # Add language-specific spacing (e.g. narrow space before : ; ? ! in French)
        try:
            text = re.sub(
                lang_data['regex'],
                lang_data[space],
                text
            )
        except re.error as e:
            raise ValueError(f"Error in regex for language {language}: {e}")

    return text


def and_list(
    elements: list,
    et: str = 'and'
) -> str:
    """Create a human-readable list, including "and"
    (or any similar word, i.e. in another language)

    :param elements: list of items to join
    :param et: conjunction word (default ``'and'``)
    :return: formatted string like ``'a, b and c'``
    """

    elements = [str(el) for el in elements]
    joined = ', '.join(elements)
    # Replace the last comma with the conjunction word (e.g. "and")
    last_comma = joined.rfind(', ')
    if last_comma == -1:
        return joined
    return joined[:last_comma] + f' {et} ' + joined[last_comma + 2:]


NAME_PREFIXES = (
    'van der',
    'von der',
    'de la',
    'del',
    'de',
    'du',
    'la',
    'von',
    'van',
    'der',
)


def capitalize_name(
    text: str
) -> str:
    """Capitalize name

    This is specially handy for names which otherwise
    would be capitalized wrongly with string's "title"
    method like:

       - Conan McArthur
       - Susanne Mayr-Grünwald
       - Maria de Angelis

    :param text: name string to capitalize
    :return: properly capitalized name
    """
    return normalize_name(
        name=text,
        returning='str'
    ) # type: ignore[return-value]  # guaranteed to be str when returning='str'


def could_be_a_name(
    name: Optional[str] = None,
    default: bool = True,
    prename: bool = False,
    lastname: bool = False
) -> bool:
    """
    Checks if a string is possibly a name

    name is preferably a full name, if not, either 'prename' or 'lastname' must be set to True

    Checks:

    - is alpha (only alpha characters)
    - there are no two consecutive uppercase characters
    - last character in any word is not uppercase
    - gender can be determined for prename or middlename

    :param name: the string to check
    :param default: the default boolean value returned if no check was successful
                    (neither positively nor negatively)
    :param prename: if True, a prename is acceptable (otherwise the full name is needed)
    :param lastname: if True, a last name is acceptable (otherwise the full name is needed)
    :return: True if the string could be a name
    """
    if prename and lastname:
        raise ValueError('prename and lastname can not be both True')

    if name is None:
        return False

    if not name.replace(
        ' ', ''
    ).replace(
        '-', ''
    ).isalpha():
        return False

    # 2 consecutive uppercase chars
    if re.search('[A-Z]{2}', name):
        return False

    human_name = cast(HumanName, normalize_name(
        name=name,
        lastname=lastname,
        returning='HumanName'
    ))

    first = human_name.first
    middle = human_name.middle
    last = human_name.last

    if prename and not first:
        return False
    if lastname and not last:
        return False

    for n in (first, middle, last):
        if not n:
            continue
        # last char is capital
        if n[-1].isupper():
            return False

        # more than 2 capitals per word
        n_words = len(n.split())
        if sum(1 for c in n if c.isupper()) > n_words + 1:
            return False

    for n in (first, middle):
        if not n:
            continue

        if get_gender(prename=n) != 'unknown':
            return True

        sum_upper = sum(1 for c in n if c.isupper())
        if sum_upper != len(n) and sum_upper > 1:
            return False

    return default


def get_gender(
    prename: str
) -> Literal['male', 'female', 'unknown']:
    """Get the gender from a prename

    :param prename: first name (compound names like ``'Anna-Maria'`` are supported)
    :return: ``'male'``, ``'female'`` or ``'unknown'``
    """
    prenames = prename.replace('-', ' ').split()

    for n in prenames:
        result = GENDER_DETECTOR.get_gender(n)
        if result in ('male', 'mostly_male'):
            return 'male'
        if result in ('female', 'mostly_female'):
            return 'female'

    return 'unknown'


def leet(
    text: str,
    max_length: Optional[int] = None,
    change_uppercase: int = 4,
    symbol_chance: int = 0,
    start_at_begin: bool = True
) -> str:
    """Leet a string

    Convert any string to leet speak.

    :param text: input text to convert
    :param max_length: optional maximum length of the returned string
    :param change_uppercase: 1 out of n chance that a character will be uppercased (default 4 = 25%)
    :param symbol_chance: 1 out of n chance that a random symbol will be added after a character
                          (0 = no symbols, default)
    :param start_at_begin: if True, truncate from the end; if False, pick a random substring
    :return: leet-speak version of the input text
    """

    text = unidecode.unidecode(text).lower()  # remove Umlaute

    leeted_text = ''

    for c in text:
        if not c.isalpha():
            continue  # without counting

        c = random.choice(CHAR_NB_MAP[c])

        if not random.randrange(change_uppercase):
            c = c.upper()

        leeted_text += c

        if symbol_chance and not random.randrange(symbol_chance):
            leeted_text += random.choice(punctuation)

    if max_length and len(leeted_text) > max_length:
        if start_at_begin:
            leeted_text = leeted_text[:max_length]
        else:
            text_length = len(leeted_text)
            max_start = max(1, text_length - max_length)
            start_position = random.randint(1, max_start)
            end_position = start_position + max_length
            leeted_text = leeted_text[start_position:end_position]

    return leeted_text


def sort_names(names: List[str]) -> List[str]:
    """Sort names

    Sorts by name, prename: splits only at the last space before sorting.

    Correctly sorts names

        - with special characters, like umlauts
        - combined names
        - names with prefixes (like de or von)

    examples:

        - Susanne Mayr-Grünwald
        - Maria de Angelis
        - Bea-Regina Hofstätter
        - Rafael

    :param names: list of name strings
    :return: sorted list of names
    """

    sort_pairs = []

    for name in names:
        human_name = cast(HumanName, normalize_name(
            name=name,
            returning='HumanName'
        ))

        first = human_name.first
        last = human_name.last

        if not last:  # only prename
            last = first

        split_at = ' '
        for name_prefix in NAME_PREFIXES:
            if f' {name_prefix} ' in last:
                split_at = f' {name_prefix} '
                break

        reversed_name = split_at.join(
            name.rsplit(split_at, 1)[::-1]
        )

        sort_key = unidecode.unidecode(reversed_name).lower()
        sort_pairs.append((sort_key, name))

    sort_pairs.sort(key=lambda pair: pair[0])

    return [name for _, name in sort_pairs]


def map_special_chars(
    text: str,
    sort: bool = False
) -> str:
    """ map special characters

    .. note:: Umlauts are even recognized if they do not use the appropriate characters but
              a vowel followed by a COMBINING DIAERESIS (U+0308) character instead.

    :param text: input text
    :param sort: if False (default), German umlauts are expanded (ä → ae, ö → oe, etc.)
                 before transliteration. If True, umlauts are transliterated 1:1 (ä → a)
                 to preserve character lengths (useful for sort-key generation).
    :return: ASCII-only transliterated text
    """
    text = normalize_text(text)

    if not sort:
        for key, value in TRANS_DICT.items():
            text = text.replace(key, value)

    return unidecode.unidecode(text)


def normalize_text(
    text: str
) -> str:
    """Normalize text

    Convert decomposed Unicode characters (e.g. vowel + COMBINING DIAERESIS U+0308)
    to their precomposed NFC form (e.g. ä, ö, ü).

    :param text: input text
    :return: NFC-normalized text
    """
    return unicodedata.normalize('NFC', text)



def normalize_name(
    name: str,
    lastname: bool = False,
    returning: Literal['str', 'dict', 'HumanName'] = 'str'
) -> Union[str, Dict, HumanName]:
    """Normalize name

    This function uses the nameparser package.

    :param name: name string to normalize
    :param lastname: if True, treat a single word as last name
    :param returning:
        - ``'str'``: return the name as a string
        - ``'dict'``: return a dictionary with the name elements
        - ``'HumanName'``: return the HumanName object from nameparser
    :return: normalized name as string, dict, or HumanName object
    """
    parsed_name = HumanName(name)
    parsed_name.capitalize()

    if lastname and parsed_name.last == '':
        parsed_name.last = parsed_name.first
        parsed_name.first = ''

    if 'str' in returning:
        return str(parsed_name)
    elif 'dict' in returning:
        return parsed_name.as_dict()
    else:  # return the HumanName object
        return parsed_name


def sort_word_list(
    word_list: List[str]
) -> List[str]:
    """Sort a list of words, taking into account special characters

    :param word_list: list of words to sort
    :return: sorted list of words
    """

    sort_pairs = [
        (map_special_chars(word, sort=True).lower(), word)
        for word in word_list
    ]
    sort_pairs.sort(key=lambda pair: pair[0])

    return [word for _, word in sort_pairs]



# coding=utf-8
"""
itb_security
============
"""

__all__ = [
    'check_password_security',
    'create_key',
    'create_secure_key',
]

from os import urandom
from secrets import choice
from string import ascii_lowercase as LOWERCASE
from string import ascii_uppercase as UPPERCASE
from string import digits as DIGITS
from string import printable as PRINTABLE
from string import whitespace as WHITESPACE
from typing import List, Literal, Optional


def create_secure_key(
    bits: int = 256,
    max_len: Optional[int] = None,
    key: Optional[str] = None,
) -> str:
    """Generate a cryptographically secure hex key

    :param bits: key strength in bits (default 256)
    :param max_len: optional maximum length of the returned string
    :param key: if provided, this key is used instead of generating a new one
    :return: hex key string
    """

    if not key:
        if bits % 8 != 0:
            raise ValueError(f'bits must be a multiple of 8, got {bits}')
        key = urandom(bits // 8).hex()

    if max_len and len(key) > max_len:
        key = key[:max_len]

    return key


def create_key(
    keyLength: int = 15,
    mode: Literal['alpha', 'alphaLower', 'alphaUpper', 'alphaNum', 'alphaNumLower', 'alphaNumUpper', 'lowerChars', 'upperChars', 'numbers', 'all'] = 'alphaNum'
) -> str:
    """Generate a random key from a character set determined by mode

    :param keyLength: length of the generated key
    :param mode: character set to use:

        - ``'alpha'``: lowercase + uppercase letters
        - ``'alphaLower'``: lowercase letters only (alias for ``lowerChars``)
        - ``'alphaUpper'``: uppercase letters only (alias for ``upperChars``)
        - ``'alphaNum'`` (default): lowercase + uppercase + digits
        - ``'alphaNumLower'``: lowercase + digits
        - ``'alphaNumUpper'``: uppercase + digits
        - ``'lowerChars'``: lowercase letters only
        - ``'upperChars'``: uppercase letters only
        - ``'numbers'``: digits only
        - ``'all'``: all printable characters except whitespace and ``^``

    :return: randomly generated key string
    """
    elements = ''

    if mode == 'alpha':
        elements = LOWERCASE + UPPERCASE
    elif mode in ('alphaLower', 'lowerChars'):
        elements = LOWERCASE
    elif mode in ('alphaUpper', 'upperChars'):
        elements = UPPERCASE
    elif mode == 'alphaNum':
        elements = LOWERCASE + UPPERCASE + DIGITS
    elif mode == 'alphaNumLower':
        elements = LOWERCASE + DIGITS
    elif mode == 'alphaNumUpper':
        elements = UPPERCASE + DIGITS
    elif mode == 'numbers':
        elements = DIGITS
    elif mode == 'all':
        elements = ''.join(set(PRINTABLE) - set(WHITESPACE) - {'^'})

    key = ''.join([choice(elements) for _ in range(keyLength)])

    return key


def check_password_security(
    pw: str,
    length: int = 8,
    check_ascii: bool = True,
    check_nonalphanum: bool = False,
    check_number: bool = False,
    check_capital: bool = False
) -> List[str]:
    """Check if a password meets certain security criteria

    :param pw: the password to check
    :param length: minimum length
    :param check_ascii: require only ASCII characters (for keyboard compatibility)
    :param check_nonalphanum: require at least one non-alphanumeric character
    :param check_number: require at least one digit
    :param check_capital: require at least one uppercase letter
    :return: list of error messages (empty if all checks pass)
    """

    errors = []

    if len(pw) < length:
        errors.append(
            f'The minimum password length should be {length}'
        )

    if check_ascii and not pw.isascii():
        errors.append(
            'Only ASCII characters are allowed in the password'
        )

    if check_number and not any(c.isdigit() for c in pw):
        errors.append(
            'The password should contain at least one digit'
        )

    if check_nonalphanum and not any(not c.isalnum() for c in pw):
        errors.append(
            'The password should contain at least one non alpha-numeric character'
        )

    if check_capital and not any(c.isupper() for c in pw):
        errors.append(
            'The password should contain at least one uppercase character'
        )

    return errors
